from __future__ import absolute_import

from ..message import Message
from . import register

@register
class event_1(Message):
    ''' Define the ``EVENT`` message (revision 1) for sending events
    between a Bokeh server and clients.

    This message is currently only generated by clients.

    '''

    msgtype  = 'EVENT'
    revision = 1

    def notify_event(self, document):
        '''

        '''
        document.apply_json_event(self.content)
